/*
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package net.sourceforge.basher;

/**
 * Defines the context parameters for a performance or testing run.
 *
 * @author Johan Lindquist
 * @version $Revision$
 */
public class BasherContext
{
    private boolean awaitStartSignal = false;
    private long setupDuration = 10;
    private long runDuration = 0;
    private long cooldownDuration = 10;

    private long startCollectionFrom = 0;
    private long stopCollectionAfter = 0;

    private int initialNumberThreads = 10;
    private int maxNumberThreads = 0;
    private int threadIncrementCount = 0;
    private int threadIncrementInterval = 0;

    private int taskMinDelay = 0;
    private int taskMaxDelay = 0;

    private long markAverageInterval = 10;

    private String beanShellScriptDirectory = "src/test/basher-scripts";

    private String runIdentifier;

    private String reportDirectory = "target/basher-reports";

    private ExecutionType executionType;

    private String name;

    private String includes;

    private String excludes;

    /**
     * Retrieves the includes expression for tasks to be included in the run defined by this context.  Defined
     * as a comma separated strings of task names (fully qualified package + task name).  Wild-cards allowed.
     * e.g. An expression of <i>net.sourceforge.*</i> would retrieve all tasks under the <i>net.sourceforge</i> package
     *
     * @return The include expression
     */
    public String getIncludes()
    {
        return includes;
    }

    /**
     * Sets the includes expression for tasks to be included in the run defined by this context.  Defined
     * as a comma separated strings of task names (fully qualified package + task name).  Wild-cards allowed.
     * e.g. An expression of <i>net.sourceforge.*</i> would retrieve all tasks under the <i>net.sourceforge</i> package
     *
     * @param includes The include expression
     */
    public void setIncludes( final String includes )
    {
        this.includes = includes;
    }

    /**
     * Sets the includes expression for tasks to be included in the run defined by this context.  Defined
     * as a comma separated strings of task names (fully qualified package + task name).  Wild-cards allowed.
     * e.g. An expression of <i>net.sourceforge.*</i> would retrieve all tasks under the <i>net.sourceforge</i> package
     *
     * @return The exclude expression
     */
    public String getExcludes()
    {
        return excludes;
    }

    /**
     * Sets the includes expression for tasks to be included in the run defined by this context.  Defined
     * as a comma separated strings of task names (fully qualified package + task name).  Wild-cards allowed.
     * e.g. An expression of <i>net.sourceforge.*</i> would retrieve all tasks under the <i>net.sourceforge</i> package
     *
     * @return The exclude expression
     */
    public void setExcludes( final String excludes )
    {
        this.excludes = excludes;
    }

    /**
     * Flag indicating whether or not the run should be started automatically upon start-up or if it is to be triggered
     * by the user.<br/>
     * If not set, this defaults to false (start automatically).
     *
     * @return True if the run is to be triggered by the user, otherwise false
     */
    public boolean isAwaitStartSignal()
    {
        return awaitStartSignal;
    }

    /**
     * Flag indicating whether or not the run should be started automatically upon start-up or if it is to be triggered
     * by the user.<br/>
     * If not set, this defaults to false (start automatically).
     *
     * @param awaitStartSignal If true, the run is to be triggered by the user, otherwise false
     */
    public void setAwaitStartSignal( final boolean awaitStartSignal )
    {
        this.awaitStartSignal = awaitStartSignal;
    }

    /**
     * Retrieves the Beanshell directory that should be used for initialization scripts.  This is relative to the working
     * directory or a full path.<br/>
     * If not set, this defaults to 'src/test/basher-scripts'.
     *
     * @return The path of the Beanshell directory
     */
    public String getBeanShellScriptDirectory()
    {
        return beanShellScriptDirectory;
    }

    /**
     * Sets the Beanshell directory that should be used for initialization scripts.  This is relative to the working
     * directory or a full path.<br/>
     * If not set, this defaults to 'src/test/basher-scripts'.
     *
     * @param beanShellScriptDirectory The Beanshell directory to use
     */
    public void setBeanShellScriptDirectory( final String beanShellScriptDirectory )
    {
        this.beanShellScriptDirectory = beanShellScriptDirectory;
    }

    /**
     * Retrieves the time (in seconds) when statistics collection should begin.  This time is relative to the start of the
     * run.
     *
     * @return The time, relative from the start of the run, when collection should start
     */
    public long getStartCollectionFrom()
    {
        return startCollectionFrom;
    }

    /**
     * Sets the time (in seconds) when statistics collection should begin.  This time is relative to the start of the
     * run.
     *
     * @param startCollectionFrom The time, relative from the start of the run, when collection should start
     */
    public void setStartCollectionFrom( final long startCollectionFrom )
    {
        this.startCollectionFrom = startCollectionFrom;
    }

    /**
     * Retrieves the time (in seconds) when statistics collection should stop.  This time is relative to the start of the
     * run.
     *
     * @return The time, relative from the start of the run, when collection should stop
     */
    public long getStopCollectionAfter()
    {
        return stopCollectionAfter;
    }

    /**
     * Sets the time (in seconds) when statistics collection should stop.  This time is relative to the start of the
     * run.
     *
     * @param stopCollectionAfter The time, relative from the start of the run, when collection should stop
     */
    public void setStopCollectionAfter( final long stopCollectionAfter )
    {
        this.stopCollectionAfter = stopCollectionAfter;
    }

    /**
     * Retrieves the number of threads which the run should start with.  Defaults to 10
     *
     * @return The number of threads the run should start with
     */
    public int getInitialNumberThreads()
    {
        return initialNumberThreads;
    }

    /**
     * Sets the number of threads which the run should start with.<br/>
     * If not set, this defaults to 10
     *
     * @param initialNumberThreads The number of threads the run should start with
     */
    public void setInitialNumberThreads( final int initialNumberThreads )
    {
        this.initialNumberThreads = initialNumberThreads;
    }

    /**
     * Retrieves the maximum number of threads allowed in the run.  Defaults to 0 (unlimited).  This is mainly taken
     * into account in combination with {@link net.sourceforge.basher.BasherContext#setThreadIncrementCount(int)} and
     * {@link net.sourceforge.basher.BasherContext#setThreadIncrementInterval(int)}.
     *
     * @return The number of threads the run should start with
     */
    public int getMaxNumberThreads()
    {
        return maxNumberThreads;
    }

    /**
     * Sets the maximum number of threads allowed in the run. This is mainly taken
     * into account in combination with {@link net.sourceforge.basher.BasherContext#setThreadIncrementCount(int)} and
     * {@link net.sourceforge.basher.BasherContext#setThreadIncrementInterval(int)}.<br/>
     * If not set, this defaults to 0 (unlimited).
     *
     * @param maxNumberThreads The number of threads the run should start with
     */
    public void setMaxNumberThreads( final int maxNumberThreads )
    {
        this.maxNumberThreads = maxNumberThreads;
    }

    /**
     * Sets the number of threads should be added for each <i>threadIncrementInterval</i> allowed in the run.<br/>
     * If not set, this defaults to 0 (don't add any)
     *
     * @return The number of threads to add
     */
    public int getThreadIncrementCount()
    {
        return threadIncrementCount;
    }

    /**
     * Retrieves the number of threads should be added for each <i>threadIncrementInterval</i> allowed in the run.<br/>
     * If not set, this defaults to 0 (don't add any)
     *
     * @param threadIncrementCount The number of threads to add
     */
    public void setThreadIncrementCount( final int threadIncrementCount )
    {
        this.threadIncrementCount = threadIncrementCount;
    }

    /**
     * Retrieves the interval (in seconds) between each attempt at adding more threads.  This will be done up till the
     * <i>maxNumberThreads</i> is reached.<br/>
     * If not set, this defaults to 0 (don't increment)
     *
     * @return The interval between adding threads (in seconds)
     */
    public int getThreadIncrementInterval()
    {
        return threadIncrementInterval;
    }

    /**
     * Sets the interval (in seconds) between each attempt at adding more threads.  This will be done, adding <i></i>
     * number of threads each time, until the <i>maxNumberThreads</i> limit is reached.<br/>
     * If not set, this defaults to 0 (don't increment)
     *
     * @param threadIncrementInterval The interval between adding threads (in seconds)
     */
    public void setThreadIncrementInterval( final int threadIncrementInterval )
    {
        this.threadIncrementInterval = threadIncrementInterval;
    }

    /**
     * Retrieves the minimum delay between running tasks.  This is mainly used to stagger the tasks.<br/>
     * If not set, this defaults to 0 (no delay).
     *
     * @return The minimum task delay (in milliseconds)
     */
    public int getTaskMinDelay()
    {
        return taskMinDelay;
    }

    /**
     * Sets the minimum delay between running tasks.  This is mainly used to stagger the tasks.<br/>
     * If not set, this defaults to 0 (no delay).
     *
     * @param taskMinDelay The maximum task delay (in milliseconds)
     */
    public void setTaskMinDelay( final int taskMinDelay )
    {
        this.taskMinDelay = taskMinDelay;
    }

    /**
     * Retrieves the maximum delay between running tasks.  This is mainly used to stagger the tasks.<br/>
     * If not set, this defaults to 0 (no delay).
     *
     * @return The maximum task delay (in milliseconds)
     */
    public int getTaskMaxDelay()
    {
        return taskMaxDelay;
    }

    /**
     * Sets the maximum delay between running tasks.  This is mainly used to stagger the tasks.<br/>
     * If not set, this defaults to 0 (no delay).
     *
     * @param taskMaxDelay The maximum task delay (in milliseconds)
     */
    public void setTaskMaxDelay( final int taskMaxDelay )
    {
        this.taskMaxDelay = taskMaxDelay;
    }

    /**
     * Retrieves the interval between average calculations (in seconds).
     *
     * @return The time between average calculations (in seconds)
     */
    public long getMarkAverageInterval()
    {
        return markAverageInterval;
    }

    /**
     * Sets the interval between average calculations (in seconds).
     *
     * @param markAverageInterval The time between average calculations (in seconds)
     */
    public void setMarkAverageInterval( final long markAverageInterval )
    {
        this.markAverageInterval = markAverageInterval;
    }


    /**
     * Retrieves the name of this Basher context.
     *
     * @return The name of the context
     */
    public String getName()
    {
        return name;
    }

    /**
     * Sets the name of this Basher context.
     *
     * @param name The name of the Basher context
     */
    public void setName( final String name )
    {
        this.name = name;
    }

    /**
     * Retrieves the report directory to be used for the running.<br/>
     * If not set, this defaults to 'target/basher-reports'.
     *
     * @return The report directory
     */
    public String getReportDirectory()
    {
        return reportDirectory;
    }

    /**
     * Sets the report directory to be used for the running.<br/>
     * If not set, this defaults to 'target/basher-reports'.
     *
     * @param reportDirectory The report directory
     */
    public void setReportDirectory( final String reportDirectory )
    {
        this.reportDirectory = reportDirectory;
    }

    /**
     * Retrieves the duration (in seconds) of the setup run.  The length defined here defines how long tasks (and the JVM) are
     * run (and how 'warmed up' the JVM could be before starting statistics collection).<br/>
     * If not set, this default to 10 seconds.
     *
     * @return The duration (in seconds) of the setup run
     */
    public long getSetupDuration()
    {
        return setupDuration;
    }

    /**
     * Sets the duration (in seconds) of the setup run.  The length defined here defines how long tasks (and the JVM) are
     * run (and how 'warmed up' the JVM could be before starting statistics collection).<br/>
     * If not set, this default to 10 seconds.
     *
     * @param setupDuration The duration (in seconds) of the setup run
     */
    public void setSetupDuration( final long setupDuration )
    {
        this.setupDuration = setupDuration;
    }

    /**
     * Retrieves the duration (in seconds) of the run.  The length defined here defines how long statistics are
     * collected for.<br/>
     *
     * @return The duration (in seconds) of the run
     */
    public long getRunDuration()
    {
        return runDuration;
    }

    /**
     * Sets the duration (in seconds) of the run.  The length defined here defines how long statistics are
     * collected for.<br/>
     *
     * @param runDuration The duration (in seconds) of the run
     */
    public void setRunDuration( final long runDuration )
    {
        this.runDuration = runDuration;
    }

    /**
     * Retrieves the duration (in seconds) of the cooldown run.  The length defined here defines how long tasks are
     * run for after stopping collection.<br/>
     * If not set, this default to 10 seconds.
     *
     * @return The duration (in seconds) of the setup run
     */
    public long getCooldownDuration()
    {
        return cooldownDuration;
    }

    /**
     * Sets the duration (in seconds) of the cooldown run.  The length defined here defines how long tasks are
     * run for after stopping collection.<br/>
     * If not set, this default to 10 seconds.
     *
     * @param cooldownDuration The duration (in seconds) of the setup run
     */
    public void setCooldownDuration( final long cooldownDuration )
    {
        this.cooldownDuration = cooldownDuration;
    }

    /**
     * @return
     */
    public ExecutionType getExecutionType()
    {
        return executionType;
    }

    /**
     * @param executionType
     */
    public void setExecutionType( final ExecutionType executionType )
    {
        this.executionType = executionType;
    }

    /**
     * @return
     */
    public String getRunIdentifier()
    {
        return runIdentifier;
    }

    /**
     * @param runIdentifier
     */
    public void setRunIdentifier( final String runIdentifier )
    {
        this.runIdentifier = runIdentifier;
    }


    @Override
    public boolean equals( final Object o )
    {
        if ( this == o )
        {
            return true;
        }
        if ( o == null || getClass() != o.getClass() )
        {
            return false;
        }

        final BasherContext that = ( BasherContext ) o;

        if ( cooldownDuration != that.cooldownDuration )
        {
            return false;
        }
        if ( initialNumberThreads != that.initialNumberThreads )
        {
            return false;
        }
        if ( markAverageInterval != that.markAverageInterval )
        {
            return false;
        }
        if ( maxNumberThreads != that.maxNumberThreads )
        {
            return false;
        }
        if ( runDuration != that.runDuration )
        {
            return false;
        }
        if ( setupDuration != that.setupDuration )
        {
            return false;
        }
        if ( startCollectionFrom != that.startCollectionFrom )
        {
            return false;
        }
        if ( stopCollectionAfter != that.stopCollectionAfter )
        {
            return false;
        }
        if ( taskMaxDelay != that.taskMaxDelay )
        {
            return false;
        }
        if ( taskMinDelay != that.taskMinDelay )
        {
            return false;
        }
        if ( threadIncrementCount != that.threadIncrementCount )
        {
            return false;
        }
        if ( threadIncrementInterval != that.threadIncrementInterval )
        {
            return false;
        }
        if ( beanShellScriptDirectory != null ? !beanShellScriptDirectory.equals( that.beanShellScriptDirectory ) : that.beanShellScriptDirectory != null )
        {
            return false;
        }
        if ( executionType != null ? !executionType.equals( that.executionType ) : that.executionType != null )
        {
            return false;
        }
        if ( name != null ? !name.equals( that.name ) : that.name != null )
        {
            return false;
        }
        if ( reportDirectory != null ? !reportDirectory.equals( that.reportDirectory ) : that.reportDirectory != null )
        {
            return false;
        }
        if ( runIdentifier != null ? !runIdentifier.equals( that.runIdentifier ) : that.runIdentifier != null )
        {
            return false;
        }

        return true;
    }

    @Override
    public int hashCode()
    {
        int result = ( int ) ( setupDuration ^ ( setupDuration >>> 32 ) );
        result = 31 * result + ( int ) ( runDuration ^ ( runDuration >>> 32 ) );
        result = 31 * result + ( int ) ( cooldownDuration ^ ( cooldownDuration >>> 32 ) );
        result = 31 * result + ( int ) ( startCollectionFrom ^ ( startCollectionFrom >>> 32 ) );
        result = 31 * result + ( int ) ( stopCollectionAfter ^ ( stopCollectionAfter >>> 32 ) );
        result = 31 * result + initialNumberThreads;
        result = 31 * result + maxNumberThreads;
        result = 31 * result + threadIncrementCount;
        result = 31 * result + threadIncrementInterval;
        result = 31 * result + taskMinDelay;
        result = 31 * result + taskMaxDelay;
        result = 31 * result + ( int ) ( markAverageInterval ^ ( markAverageInterval >>> 32 ) );
        result = 31 * result + ( beanShellScriptDirectory != null ? beanShellScriptDirectory.hashCode() : 0 );
        result = 31 * result + ( runIdentifier != null ? runIdentifier.hashCode() : 0 );
        result = 31 * result + ( reportDirectory != null ? reportDirectory.hashCode() : 0 );
        result = 31 * result + ( executionType != null ? executionType.hashCode() : 0 );
        result = 31 * result + ( name != null ? name.hashCode() : 0 );
        return result;
    }

    @Override
    public String toString()
    {
        return "BasherContext{" +
                "setupDuration=" + setupDuration +
                ", runDuration=" + runDuration +
                ", cooldownDuration=" + cooldownDuration +
                ", startCollectionFrom=" + startCollectionFrom +
                ", stopCollectionAfter=" + stopCollectionAfter +
                ", initialNumberThreads=" + initialNumberThreads +
                ", maxNumberThreads=" + maxNumberThreads +
                ", threadIncrementCount=" + threadIncrementCount +
                ", threadIncrementInterval=" + threadIncrementInterval +
                ", taskMinDelay=" + taskMinDelay +
                ", taskMaxDelay=" + taskMaxDelay +
                ", markAverageInterval=" + markAverageInterval +
                ", beanShellScriptDirectory='" + beanShellScriptDirectory + '\'' +
                ", runIdentifier='" + runIdentifier + '\'' +
                ", reportDirectory='" + reportDirectory + '\'' +
                ", executionType=" + executionType +
                ", name='" + name + '\'' +
                '}';
    }
}