package net.sourceforge.basher.impl;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import net.sourceforge.basher.Task;

/**
 * @author Johan Lindquist
 */
public class TaskTree
{
    private TaskNode _root = new TaskNode( "ROOT" );

    public TaskTree( final List<Task> availableTasksWorkingList )
    {
        for ( Task task : availableTasksWorkingList )
        {
            addTask( task );
        }
    }

    public TaskTree()
    {
    }

    private class TaskNode
    {
        private String _identifier;
        private List<Task> _tasks;
        private List<TaskNode> _branches;

        private TaskNode( final String identifier )
        {
            _identifier = identifier;
        }

        public TaskNode getBranch( final String name )
        {
            if ( _branches == null )
            {
                return null;
            }

            for ( TaskNode branch : _branches )
            {
                if ( branch._identifier.equals( name ) )
                {
                    return branch;
                }
            }
            return null;
        }

        public List<Task> getTasks()
        {
            if ( _tasks == null )
            {
                return Collections.emptyList();
            }
            return _tasks;
        }

        public List<Task> getAllTasks()
        {
            List<Task> allTasks = new ArrayList<Task>( (_tasks == null ? Collections.<Task>emptyList() : _tasks));
            if (_branches == null)
            {
                return allTasks;
            }
            for ( TaskNode branch : _branches )
            {
                allTasks.addAll( branch.getAllTasks() );
            }
            return allTasks;
        }

        public void addBranch( final TaskNode taskNode )
        {
            if ( _branches == null )
            {
                _branches = new ArrayList<TaskNode>();
            }

            // Check if the branch exists already
            for ( TaskNode branch : _branches )
            {
                if ( branch._identifier.equals( taskNode._identifier ) )
                {
                    return;
                }
            }
            _branches.add( taskNode );
        }

        public void addTask( final Task task )
        {
            if ( _tasks == null )
            {
                _tasks = new ArrayList<Task>();
            }
            _tasks.add( task );
        }
    }

    public List<Task> getTasks( final String pattern )
    {
        // Pattern is something like
        // net.sourceforge.basher.**.*
        //                        ^  ^
        //                        |  '- Wildcard for a set of tasks within the package
        //                        |
        //                        '- Wildcard for a set of packages

        String[] splitBranches = pattern.split( "\\." );

        TaskNode taskNode = _root;
        for ( int i = 0; i < splitBranches.length-1; i++ )
        {
            String branch = splitBranches[ i ];

            // If pattern == wildcard (**) this should evaluate all
            // branches at this level

            final TaskNode nextTaskNode = taskNode.getBranch( branch );
            if ( nextTaskNode == null )
            {
                // No more splitBranches - return tasks in this node
                return taskNode.getTasks();
            }
            else
            {
                taskNode = nextTaskNode;
            }

        }
        return taskNode.getAllTasks();


    }

    public void addTask( Task task )
    {
        String taskName = task.getName();
        String[] branches = taskName.split( "\\." );

        TaskNode taskNode = _root;

        for ( int i = 0; i < branches.length - 1; i++ )
        {
            String branch = branches[ i ];
            TaskNode node = taskNode.getBranch( branch );
            if ( node == null )
            {
                final TaskNode newNode = new TaskNode( branch );
                taskNode.addBranch( newNode );
                node = newNode;
            }
            taskNode = node;
        }
        taskNode.addTask( task );

    }

}
