package net.sourceforge.basher.internal.impl;

import java.lang.reflect.Method;
import java.lang.reflect.Constructor;
import java.lang.reflect.Modifier;

import net.sourceforge.basher.Task;
import net.sourceforge.basher.BasherException;
import net.sourceforge.basher.Phase;
import net.sourceforge.basher.annotations.*;
import net.sourceforge.basher.internal.TaskDecorator;
import org.ops4j.gaderian.service.*;
import org.apache.commons.logging.Log;

/**
 * @author Johan Lindquist
 * @version $Revision$
 */
public class TaskDecoratorImpl implements TaskDecorator
{
    private ClassFactory _classFactory;
    private Log _log;

    public void setLog( final Log log )
    {
        _log = log;
    }

    public void setClassFactory( final ClassFactory classFactory )
    {
        _classFactory = classFactory;
    }

    public Task decorateInstance( final Object taskInstance )
    {
        try
        {
            // TODO: Validation - instance must be decorated with annotations or implement Task

            final String taskInstanceClassName = taskInstance.getClass().getName();
            final String className = taskInstanceClassName.substring( taskInstanceClassName.lastIndexOf( '.' ) + 1 );

            final String taskClassName = className + "BasherDecoratedTask";

            // Determine the execution method name
            final String executionMethodName = determineExecutionMethodName( taskInstance );

            // Then create the decorated task
            final DecoratedTask decoratedTask = createDecoratedTask( taskClassName, executionMethodName, taskInstance );

            // Pass the instance to check for annotations and initialize the task
            initializeTask( taskInstance, decoratedTask );

            return decoratedTask;
        }
        catch ( Exception e )
        {
            throw new BasherException( e.getMessage(), e );
        }
    }

    private void initializeTask( final Object taskInstance, final DecoratedTask decoratedTask )
    {
        final Class<? extends Object> taskInstanceClass = taskInstance.getClass();

        // First initialize from annotations
        processAnnotations( taskInstanceClass, decoratedTask );

        // FIXME: This should auto-wire any task dependencies

    }

    private void processMethods( final Class<? extends Object> taskInstanceClass, final ClassFab decoratedTask )
    {
        if ( taskInstanceClass.isAssignableFrom( Task.class ) )
        {
            // No need to 'enhance' this - already implements the class
            return;
        }

        // By doing depth-first here, we allow sub-classes to override super classes methods
        final Class superClass = taskInstanceClass.getSuperclass();
        if ( !superClass.equals( Object.class ) )
        {
            processMethods( superClass, decoratedTask );
        }


        final Method[] taskMethods = Task.class.getDeclaredMethods();
        for ( int i = 0; i < taskMethods.length; i++ )
        {
            Method taskMethod = taskMethods[ i ];

            try
            {
                final Method declaredMethod = taskInstanceClass.getDeclaredMethod( taskMethod.getName(), taskMethod.getParameterTypes() );

                final MethodSignature methodSignature = new MethodSignature( declaredMethod );

                final MethodFab methodFab = decoratedTask.getMethodFab( methodSignature );

                // Only if we haven't enhanced this method already do we do something
                if ( methodFab == null )
                {
                    // If we get here, it means we have the method available in the task
                    // If so, add method, proxying the call to the instance
                    decoratedTask.addMethod( Modifier.PUBLIC, methodSignature, "return _taskInstance." + declaredMethod.getName() + "();" );
                }

            }
            catch ( NoSuchMethodException e )
            {
                // Safe to ignore - class may not actually declare it
            }

        }

    }

    private void processAnnotations( final Class<? extends Object> taskInstanceClass, final DecoratedTask decoratedTask )
    {
        // First examine super classes for annotations
        // By doing depth-first here, we allow sub-classes to override super classes annotations
        final Class superClass = taskInstanceClass.getSuperclass();
        if ( !superClass.equals( Object.class ) )
        {
            processAnnotations( superClass, decoratedTask );
        }

        // Determine whether the task is timed or not
        if (taskInstanceClass.isAnnotationPresent( BasherTimedTask.class ))
        {
            BasherTimedTask annotation = taskInstanceClass.getAnnotation( BasherTimedTask.class );
            decoratedTask.setInterval( annotation.interval() );
            decoratedTask.setTimed( true );
        }

        if ( taskInstanceClass.isAnnotationPresent( BasherMaxTime.class ) )
        {
            BasherMaxTime annotation = taskInstanceClass.getAnnotation( BasherMaxTime.class );
            decoratedTask.setMaxTime( annotation.time() );
        }
        if ( taskInstanceClass.isAnnotationPresent( BasherMaxInvocations.class ) )
        {
            BasherMaxInvocations annotation = taskInstanceClass.getAnnotation( BasherMaxInvocations.class );
            decoratedTask.setMaxInvocations( annotation.max() );
        }
        if ( taskInstanceClass.isAnnotationPresent( BasherInertia.class ) )
        {
            BasherInertia annotation = taskInstanceClass.getAnnotation( BasherInertia.class );
            decoratedTask.setInertia( annotation.inertia() );
        }
        if ( taskInstanceClass.isAnnotationPresent( BasherName.class ) )
        {
            BasherName annotation = taskInstanceClass.getAnnotation( BasherName.class );
            decoratedTask.setName( ( ( BasherName ) annotation ).name() );
        }
        if ( taskInstanceClass.isAnnotationPresent( BasherStopAfter.class ) )
        {
            BasherStopAfter annotation = taskInstanceClass.getAnnotation( BasherStopAfter.class );
            decoratedTask.setStopAfter( ( ( BasherStopAfter ) annotation ).time() );
        }
        if ( taskInstanceClass.isAnnotationPresent( BasherWeight.class ) )
        {
            BasherWeight annotation = taskInstanceClass.getAnnotation( BasherWeight.class );
            decoratedTask.setWeight( ( ( BasherWeight ) annotation ).weight() );
        }
        if ( taskInstanceClass.isAnnotationPresent( BasherRunFrom.class ) )
        {
            BasherRunFrom annotation = taskInstanceClass.getAnnotation( BasherRunFrom.class );
            decoratedTask.setRunFrom( ( ( BasherRunFrom ) annotation ).time() );
        }

        if ( taskInstanceClass.isAnnotationPresent( BasherPhases.class ) )
        {
            decoratedTask.clearPhases();
            Phase[] phases = taskInstanceClass.getAnnotation( BasherPhases.class ).phases();
            for ( Phase phase : phases )
            {
                decoratedTask.addPhase( phase );
            }
        }

        if ( decoratedTask.applicablePhases().size() == 0 )
        {
            _log.warn( "No applicable phase specified for task" );
        }

    }

    private DecoratedTask createDecoratedTask( final String decoratedTaskClassName, final String executionMethodName, final Object taskInstance ) throws Exception
    {
        final ClassFab fab = _classFactory.newClass( decoratedTaskClassName, DecoratedTask.class );
        fab.addField( "_taskInstance", taskInstance.getClass() );

        final BodyBuilder bodyBuilder = new BodyBuilder();
        bodyBuilder.begin();
        bodyBuilder.addln( "_taskInstance." + executionMethodName + "();" );
        bodyBuilder.end();

        fab.addMethod( java.lang.reflect.Modifier.PUBLIC, new MethodSignature( void.class, "doExecuteTask", new Class[0], new Class[]{ Throwable.class } ), bodyBuilder.toString() );

        bodyBuilder.clear();

        // Now, add override for all 'Task' methods
        processMethods( taskInstance.getClass(), fab );

        bodyBuilder.begin();
        bodyBuilder.addln( "super();" );
        bodyBuilder.addln( "_taskInstance = $1;" );
        bodyBuilder.end();

        fab.addConstructor( new Class[]{ taskInstance.getClass() }, new Class[0], bodyBuilder.toString() );

        final Class decoratedTask = fab.createClass();

        final Constructor constructor = decoratedTask.getConstructor( taskInstance.getClass() );

        return ( DecoratedTask ) constructor.newInstance( taskInstance );

    }

    String determineExecutionMethodName( final Object taskInstance )
    {
        final Method[] methods = taskInstance.getClass().getMethods();

        Method selectedMethod = null;

        for ( final Method method : methods )
        {
            if ( method.getAnnotation( BasherExecuteMethod.class ) != null )
            {
                if ( selectedMethod == null )
                {
                    // Found an execution method
                    selectedMethod = method;
                }
                else
                {
                    throw new BasherException( "Found more than 1 execute method", null );
                }
            }
        }

        if ( selectedMethod != null )
        {
            validateExecutionMethod( selectedMethod );
            return selectedMethod.getName();
        }

        // Ok, look for the default method.
        try
        {
            taskInstance.getClass().getMethod( "executeTask" );
            return "executeTask";
        }
        catch ( NoSuchMethodException e )
        {
            throw new BasherException( "Could not find executeTask", e );
        }

    }

    private void validateExecutionMethod( Method method )
    {
        if ( method.getParameterTypes().length != 0 )
        {
            throw new BasherException( "Execution method '" + method.getName() + "'  must not take any parameters", null );
        }

    }
}
