/*
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package net.sourceforge.basher.internal.tasks;

import java.util.TimerTask;

import net.sourceforge.basher.BasherContext;
import net.sourceforge.basher.ContextManager;
import net.sourceforge.basher.Scheduler;
import net.sourceforge.basher.events.BasherEvent;
import net.sourceforge.basher.events.BasherEventListener;
import net.sourceforge.basher.events.PhaseTransitionEvent;
import net.sourceforge.basher.tasks.AbstractTask;
import org.apache.commons.logging.Log;

/**
 * @author Johan Lindquist
 * @version 1.0
 */
public class ThreadIncrementTask extends TimerTask
{
    private Log _log;
    private Scheduler _scheduler;
    private ContextManager _contextManager;
    private boolean _skipIncrementing = false;

    public void setContextManager( final ContextManager contextManager )
    {
        _contextManager = contextManager;
    }

    public void setLog( final Log log )
    {
        _log = log;
    }

    public void setScheduler(final Scheduler scheduler)
    {
        _scheduler = scheduler;
    }

    @Override
    public void run()
    {
        try
        {
            final BasherContext basherContext = _contextManager.getActiveBasherContext();
            int threadIncrementCount = basherContext.getThreadIncrementCount();

            if (threadIncrementCount == 0)
            {
                _log.warn("Thread increment of '0' thread(s)");
                return;
            }

            if (_scheduler.getNumberOfActiveThreads() == basherContext.getMaxNumberThreads() && !_skipIncrementing)
            {
                _log.info( "Maximum threads reached, not incrementing further" );
                _skipIncrementing = true;
                return;
            }

            if (_skipIncrementing)
            {
                // Simply return
                return;
            }


            // Reset this to false in case thread count has dropped
            _skipIncrementing = false;

            _log.info("Incrementing thread count by '" + threadIncrementCount + "' thread(s)");
            _scheduler.addThreads(threadIncrementCount);
            _log.info("Number of threads incremented by '" + threadIncrementCount + "' thread(s)");

        }
        catch (Throwable e)
        {
            _log.error(e.getMessage(), e);
        }

    }
}
